#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <ctype.h>


unsigned long CalcTableChecksum(unsigned long* table, unsigned long length);

struct name_record {
	unsigned short		platform;
	unsigned short		encoding;
	unsigned short		language;
	unsigned short		name;
	char*				string;
	size_t				string_length;
};


main(int argc, char* argv[])
{
	if (argc != 3) {
		printf("Usage: setfontname filename fontname\n");
		exit(2);
	}

	char*			font_name = argv[2];
	unsigned short	font_name_length = strlen(font_name);

	char*			file_name = argv[1];
	FILE*			in_file = fopen(file_name, "rt");
	if (in_file == NULL) {
		printf("\"%s\" not found\n", file_name);
		exit(2);
	}
	struct	stat	info;
	fstat(fileno(in_file), &info);
	size_t			buffer_size = info.st_size;
	char*			buffer = (char*) malloc(buffer_size);
	fread(buffer, sizeof(char), buffer_size, in_file);
	fclose(in_file);


	unsigned short	table_count = *(unsigned short *) &buffer[4];
	unsigned long	name_offset, name_length = 0, name_header_offset;
	for (unsigned short i = 0; i < table_count; i++)
		if (*(unsigned long*) &buffer[0x0C + i * 0x10] == 'name') {
			name_offset = *(unsigned long*) &buffer[0x0C + 0x08 + i * 0x10];
			name_length = *(unsigned long*) &buffer[0x0C + 0x0C + i * 0x10];
			name_header_offset = 0x0C + i * 0x10;
			break;
		}
	if (name_length == 0) {
		printf("'name' tag not found\n");
		exit(2);
	}




	unsigned short		record_count = *(unsigned short*) &buffer[name_offset + 0x02];
	unsigned long		string_offset = name_offset + *(unsigned short*) &buffer[name_offset + 0x04];
	name_record*		record_array = (name_record*) malloc(record_count * sizeof(name_record));
	for (unsigned short i = 0; i < record_count; i++) {
		unsigned short*		data = (unsigned short*) &buffer[name_offset + 0x06 + i * 0x0C];
		record_array[i].platform = *data++;
		record_array[i].encoding = *data++;
		record_array[i].language = *data++;
		record_array[i].name = *data++;
		record_array[i].string_length = *data++;
		record_array[i].string = (char*) malloc(record_array[i].string_length);
		memcpy(record_array[i].string, buffer + string_offset + *data++,
			record_array[i].string_length);
	}
	
	unsigned long		new_name_length = 0x06;
	for (unsigned short i = 0; i < record_count; i++)
		if ((record_array[i].platform != 3 || record_array[i].encoding != 1) && record_array[i].name == 1)
			new_name_length += 0x0C + font_name_length;
		else
			new_name_length += 0x0C + record_array[i].string_length;

	size_t		new_buffer_size = buffer_size - ((name_length + 3) & ~3) + ((new_name_length + 3) & ~3);
	char*		new_buffer = (char*) malloc(new_buffer_size);
	memcpy(new_buffer, buffer, name_offset);
	memcpy(new_buffer + name_offset + ((new_name_length + 3) & ~3),
			buffer + name_offset + ((name_length + 3) & ~3),
			buffer_size - name_offset - ((name_length + 3) & ~3));

	string_offset = name_offset + 0x06 + 0x0C * record_count;
	*(unsigned short*) &new_buffer[name_offset + 0x00] = 0;
	*(unsigned short*) &new_buffer[name_offset + 0x02] = record_count;
	*(unsigned short*) &new_buffer[name_offset + 0x04] = string_offset - name_offset;
	unsigned short		current_string_offset = 0;
	for (unsigned short i = 0; i < record_count; i++) {
		unsigned short*		data = (unsigned short*) &new_buffer[name_offset + 0x06 + i * 0x0C];
		*data++ = record_array[i].platform;
		*data++ = record_array[i].encoding;
		*data++ = record_array[i].language;
		*data++ = record_array[i].name;
		unsigned short		string_length = record_array[i].string_length;
		if ((record_array[i].platform != 3 || record_array[i].encoding != 1) && record_array[i].name == 1)
			string_length = font_name_length;
		*data++ = string_length;
		*data++ = current_string_offset;
		current_string_offset += string_length;
	}
	char*	string_buffer = new_buffer + string_offset;
	for (unsigned short i = 0; i < record_count; i++)
		if ((record_array[i].platform != 3 || record_array[i].encoding != 1) && record_array[i].name == 1) {
			memcpy(string_buffer, font_name, font_name_length);
			string_buffer += font_name_length;
		} else {
			memcpy(string_buffer, record_array[i].string, record_array[i].string_length);
			string_buffer += record_array[i].string_length;
		}
	unsigned long	pad = 0;
	memcpy(string_buffer, &pad, ((new_name_length + 3) & ~3) - new_name_length);
	for (unsigned short i = 0; i < record_count; i++)
		free(record_array[i].string);
	free(record_array);
	free(buffer);


	for (unsigned short i = 0; i < table_count; i++) {
		if (*(unsigned long*) &new_buffer[0x0C + 0x00 + i * 0x10] == 'name') {
			*(unsigned long*) &new_buffer[0x0C + 0x0C + i * 0x10] = new_name_length;
			*(unsigned long*) &new_buffer[0x0C + 0x04 + i * 0x10] = CalcTableChecksum((unsigned long*) (new_buffer + name_offset), new_name_length);;
		}
		if (*(unsigned long*) &new_buffer[0x0C + 0x08 + i * 0x10] > name_offset)
			*(unsigned long*) &new_buffer[0x0C + 0x08 + i * 0x10] += ((new_name_length + 3) & ~3) - ((name_length + 3) & ~3);
	}

	FILE*	out_file = fopen(file_name, "wt");
	fwrite(new_buffer, sizeof(char), new_buffer_size, out_file);
	fclose(out_file);
	free(new_buffer);
}

unsigned long
CalcTableChecksum(unsigned long* table, unsigned long length)
{
	unsigned long	sum = 0L;
	unsigned long*	end_ptr = table + ((length + 3) & ~3) / sizeof(unsigned long);
	while (table < end_ptr)
		sum += *table++;
	return sum;
}

